# coding=utf-8
import random

import cv2
import keyboard
import numpy as np

import requests
import vision_definitions
from naoqi import ALProxy

from movement import NaoMotion


# 获取行走配置

def init_config():
    maxstepx = 0.066
    # maxstepx = 0.33
    maxstepy = 0.14
    # maxstepy = 0.22
    maxsteptheta = 0.4
    # maxsteptheta = 0.45
    maxstepfrequency = 0.7
    # maxstepfrequency = 1.8
    stepheight = 0.02
    # stepheight = 0.04
    torsowx = 0.0
    torsowy = 0.0
    return [["MaxStepX", maxstepx],
            ["MaxStepY", maxstepy],
            ["MaxStepTheta", maxsteptheta],
            ["MaxStepFrequency", maxstepfrequency],
            ["StepHeight", stepheight],
            ["TorsoWx", torsowx],
            ["TorsoWy", torsowy]]


def get_move_config():
    # maxstepx = 0.066
    maxstepx = 0.33
    # maxstepy = 0.14
    maxstepy = 0.22
    # maxsteptheta = 0.4
    maxsteptheta = 0.45
    # maxstepfrequency = 0.7
    maxstepfrequency = 1.8
    # stepheight = 0.02
    stepheight = 0.04
    torsowx = 0.0
    torsowy = 0.0
    return [["MaxStepX", maxstepx],
            ["MaxStepY", maxstepy],
            ["MaxStepTheta", maxsteptheta],
            ["MaxStepFrequency", maxstepfrequency],
            ["StepHeight", stepheight],
            ["TorsoWx", torsowx],
            ["TorsoWy", torsowy]]


# 图像处理逻辑的集成
def find_white_lines(nao_image):
    # 转换为HSV色彩空间并应用色彩阈值
    hsv_image = cv2.cvtColor(nao_image, cv2.COLOR_BGR2HSV)
    lower_threshold = np.array([0, 0, 100])
    upper_threshold = np.array([180, 50, 255])
    mask = cv2.inRange(hsv_image, lower_threshold, upper_threshold)

    # 应用形态学操作
    kernel = np.ones((3, 3), np.uint8)
    mask = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, kernel)
    mask = cv2.morphologyEx(mask, cv2.MORPH_OPEN, kernel)

    # 应用霍夫变换来检测线条
    Lines = cv2.HoughLines(mask, 1, np.pi / 260, 80)
    return Lines


# 获取视频配置
def get_video_config():
    resolution = vision_definitions.kQVGA
    colorSpace = vision_definitions.kBGRColorSpace
    fps = 5
    return [resolution, colorSpace, fps]


# 判断线条是否与图像上下边缘近似平行
def is_line_parallel(rho, theta):
    a = np.cos(theta)
    b = np.sin(theta)
    x0 = a * rho
    y0 = b * rho
    x1 = int(x0 + 1000 * (-b))
    y1 = int(y0 + 1000 * a)
    x2 = int(x0 - 1000 * (-b))
    y2 = int(y0 - 1000 * a)

    # 检查线条是否与图像上下边缘近似平行
    # 这里检查线条的斜率（注意处理除以零的情况）
    if x2 - x1 != 0:
        slope = (y2 - y1) / (x2 - x1)
        return abs(slope) < 1  # some_threshold 是您定义的阈值
    else:
        return False


class NaoRobot:
    def __init__(self, NAO_IP, NAO_PORT):
        self.arms_thread = None
        self.motion = ALProxy("ALMotion", NAO_IP, NAO_PORT)
        self.video_service = ALProxy("ALVideoDevice", NAO_IP, NAO_PORT)
        self.tts = ALProxy("ALTextToSpeech", NAO_IP, NAO_PORT)
        self.posture = ALProxy("ALRobotPosture", NAO_IP, NAO_PORT)
        self.memory = ALProxy("ALMemory", NAO_IP, NAO_PORT)
        self.life = ALProxy("ALAutonomousLife", NAO_IP, NAO_PORT)
        self.sound = ALProxy("ALAudioSourceLocalization", NAO_IP, NAO_PORT)
        self.audio = ALProxy("ALAudioDevice", NAO_IP, NAO_PORT)
        self.moveConfig = get_move_config()
        self.init_config = init_config()
        self.server_url = "http://192.168.30.243:5000/"
        self.crossed_finish_line = False  # 添加一个标志来表示是否越过终点线
        self.rearEng = None
        self.frontEng = None
        self.leftEng = None
        self.rightEng = None
        self.initial_position = self.motion.getRobotPosition(True)  # 记录初始位置
        self.nao_motion = NaoMotion(NAO_IP, NAO_PORT)

    # 等待头部被触摸
    def wait_for_head_touch(self):
        print("Waiting for head touch...")
        while True:
            touch_value = self.memory.getData("Device/SubDeviceList/Head/Touch/Middle/Sensor/Value")
            if touch_value > 0.5:
                print("Head touched!")
                break

    # 进入预备姿势
    def prepare_for_walk(self):
        self.wait_for_head_touch()
        self.start_arms_pose()
        # 这里可以继续添加其他准备动作

    def get(self):
        self.audio.enableEnergyComputation()
        self.leftEng = self.audio.getLeftMicEnergy()
        self.rightEng = self.audio.getRightMicEnergy()
        self.frontEng = self.audio.getFrontMicEnergy()
        self.rearEng = self.audio.getRearMicEnergy()
        return self.leftEng + self.rightEng + self.frontEng + self.rearEng

    # 哨声检测逻辑
    def check_sound_energy(self):
        print("Listening for sound...")
        while True:
            # 获取当前的声音能量值
            sound_energy = self.get()

            # 判断声音能量是否超过阈值
            if sound_energy > 40000:
                print("High sound energy detected!")
                return True
            else:
                print("Low sound energy detected.")

    # 创建一个线程来执行特定动作
    def start_arms_pose(self):
        # 让机器人保持低头
        self.posture.goToPosture("StandInit", 0.5)
        # 让机器人的手指紧握
        self.motion.setAngles("LHand", 0, 0.2)

    # 获取图片
    def process_image(self, VideoClient):
        nao_image = self.video_service.getImageRemote(VideoClient)
        # print(nao_image)

        if nao_image is None:
            return None, None

        image_width = nao_image[0]
        image_height = nao_image[1]
        array = nao_image[6]

        Img = np.frombuffer(array, dtype=np.uint8).reshape((image_height, image_width, 3))
        # print(Img)
        White_lines = find_white_lines(Img)
        if White_lines is not None:
            for line in White_lines:
                rho, theta = line[0]
                if is_line_parallel(rho, theta):
                    self.crossed_finish_line = True
                    break
                else:
                    self.crossed_finish_line = False
        else:
            print("No white lines detected.")

        return Img, White_lines

    # 添加一个方法来检查是否越过了终点线
    def check_if_arrived(self):
        # 设置总共行走的距离，如果已经走了固定的距离，就认为已经到达了终点
        total_distance = 6.5
        # 获取当前的位置
        current_position = self.motion.getRobotPosition(True)
        if current_position - self.initial_position >= total_distance:
            print("Arrived!")
            return True
        # print("Not arrived yet.")
        return False

    # 根据检测到的线条调整运动
    def adjust_motion(self, detected_lines):
        # 如果检测到线条
        if detected_lines is not None and len(detected_lines) > 0:
            single_line = detected_lines[0]
            rho, theta = single_line[0]

            rotation_value = 0.05  # 旋转的基本值
            if theta <= 1.04 or 1.57 < theta <= 2.09:
                # 向右转
                self.motion.move(0.1, 0.0, -rotation_value, self.moveConfig)
            else:
                # 向左转
                self.motion.move(0.1, 0.0, rotation_value, self.moveConfig)
        else:
            # 如果没有检测到线条，直接前进
            self.motion.move(0.8, 0.0, 0.0, self.moveConfig)

    # 向服务器发送信号
    def send_relay_signal(self):
        try:
            response = requests.get(self.server_url + "startSecondLeg")
            print(response.status_code)
            if response.status_code == 200:
                print("Signal sent successfully.")
            else:
                print("Failed to send signal, status code:", response.status_code)
        except requests.RequestException as e:
            print("Error sending signal:", e)


if __name__ == '__main__':
    IP = "192.168.30.11"
    # IP = "192.168.30.242"
    PORT = 9559

    # 创建一个NaoRobot对象
    nao_robot = NaoRobot(IP, PORT)
    videoConfig = get_video_config()

    # 关闭自主模式
    nao_robot.life.setState("disabled")

    nao_robot.prepare_for_walk()
    nao_robot.check_sound_energy()

    videoClient = nao_robot.video_service.subscribeCamera("python_nao" + str(random.randint(1, 100000)), 1,
                                                          videoConfig[0], videoConfig[1],
                                                          videoConfig[2])
    nao_robot.video_service.setCameraParameter(videoClient, 22, 2)

    try:
        while True:
            img, lines = nao_robot.process_image(videoClient)
            # print("lines={}".format(lines))
            # print("img={}".format(img))

            # 如果图像不为空
            if img is not None:
                cv2.imshow("NAO Camera", img)
                # 按q退出
                if cv2.waitKey(1) & 0xFF == ord('q'):
                    # 停止当前所有行为
                    nao_robot.motion.stopMove()
                    nao_robot.posture.goToPosture("StandInit", 0.5)
                    break
                # 调整运动
                nao_robot.adjust_motion(lines)

                # # 检查是否到达终点
                # if nao_robot.check_if_arrived():
                #     break
                # else:
                #     pass
            else:
                print("No image received.")
    finally:
        # 通过键盘控制机器人转向和移动
        while True:
            # 获取键盘输入
            key = cv2.waitKey(1) & 0xFF
            # 按q退出
            if key == ord('q'):
                break
            # 按左键转向
            elif key == ord('a'):
                nao_robot.motion.move(0.0, 0.0, 0.1, nao_robot.init_config)
            # 按右键转向
            elif key == ord('d'):
                nao_robot.motion.move(0.0, 0.0, -0.1, nao_robot.init_config)
            # 按上键前进
            elif key == ord('w'):
                nao_robot.motion.move(0.1, 0.0, 0.0, nao_robot.init_config)
            # 按下键后退
            elif key == ord('s'):
                nao_robot.motion.move(-0.1, 0.0, 0.0, nao_robot.init_config)
            # 按z紧握
            elif key == ord('z'):
                nao_robot.motion.setAngles("LHand", 0, 0.5)
                # nao_robot.motion.setAngles("RHand", 0, 0.5)
            # 按x松手
            elif key == ord('x'):
                nao_robot.motion.setAngles("LHand", 1, 0.5)
                # nao_robot.motion.setAngles("RHand", 1, 0.5)
            # 按空格键停止
            elif key == ord(' '):
                nao_robot.motion.stopMove()
            # 按p双腿直立
            elif key == ord('p'):
                nao_robot.posture.goToPosture("Stand", 1.0)
                nao_robot.motion.setAngles("LHand", 0, 0.5)
            # 按c进入交棒姿势
            elif key == ord('c'):
                nao_robot.motion.setAngles("LHand", 0, 0.5)
                nao_robot.nao_motion.set_first_arms()

            elif keyboard.is_pressed('left arrow'):
                # 向左移动
                nao_robot.motion.move(0, 0.1, 0, nao_robot.init_config)
                nao_robot.motion.setAngles("LHand", 0, 0.5)
            elif keyboard.is_pressed('right arrow'):
                # 向右移动
                nao_robot.motion.move(0, -0.1, 0, nao_robot.init_config)
                nao_robot.motion.setAngles("LHand", 0, 0.5)
            else:
                pass
        # print("111111111111111111111111111111111")
        nao_robot.send_relay_signal()
        nao_robot.video_service.unsubscribe(videoClient)
        print("Unsubscribed from camera.")

        cv2.destroyAllWindows()

        # 通过键盘控制机器人转向和移动
        while True:
            # 获取键盘输入
            # if keyboard.is_pressed('q'):
            #     break  # 按q退出
            if keyboard.is_pressed('a'):
                nao_robot.motion.moveToward(0, 0, 0.5)  # 按左键转向
            if keyboard.is_pressed('d'):
                nao_robot.motion.moveToward(0, 0, -0.5)  # 按右键转向
            if keyboard.is_pressed('w'):
                nao_robot.motion.moveToward(0.5, 0, 0)  # 按上键前进
            if keyboard.is_pressed('s'):
                nao_robot.motion.moveToward(-0.5, 0, 0)  # 按下键后退
            if keyboard.is_pressed('z'):
                nao_robot.motion.setAngles("LHand", 0, 0.5)  # 按z键紧握
                # nao_robot.motion.setAngles("RHand", 0, 0.5)  # 按z键紧握
            if keyboard.is_pressed('x'):
                nao_robot.motion.setAngles("LHand", 1, 0.5)  # 按x键松开
                # nao_robot.motion.setAngles("RHand", 1, 0.5)  # 按x键松开
            if keyboard.is_pressed('space'):
                nao_robot.motion.stopMove()  # 按空格键停止
            if keyboard.is_pressed('r'):
                nao_robot.posture.goToPosture("StandInit", 0.5)  # 按"r"键进入standInit姿势
            if keyboard.is_pressed('left arrow'):
                # 向左移动
                nao_robot.motion.moveToward(0, 0.5, 0)  # (x, y, theta), y > 0 表示向左
            if keyboard.is_pressed('right arrow'):
                # 向右移动
                nao_robot.motion.moveToward(0, -0.5, 0)  # y < 0 表示向右
            if keyboard.is_pressed('p'):
                nao_robot.motion.wakeUp()
                nao_robot.posture.goToPosture("Stand", 1.0)  # 双腿直立
